"""LightningFlower Default Configuration"""
from enum import Enum


class DatasetConfig(Enum):
    NON_IID = 0
    IID = 1
    SAMPLED = 2


class LightningFlowerDefaults(object):
    NR_ROUNDS = 2
    GRPC_MAX_MSG_LENGTH = 1073741824  # 1024x1024x1024
    HOST_ADDRESS = "localhost:8081"
    CLIENT_ID = 1
    WEIGHTS_FILENAME = "weights.npz"
    WEIGHTS_FOLDER = "./saved_weights"
    DATASET_FOLDER = './data'
    BATCH_SIZE_TRAIN = 32
    BATCH_SIZE_TEST = 32
    DATASET_CONFIG = DatasetConfig.IID
    NUM_WORKERS = 2
    NUM_CLIENTS = 2



"""LightningFlower Data Module"""
import numpy as np
import os
import torch.utils
from torch.utils.data import DataLoader, SubsetRandomSampler
from lightningflower.config import LightningFlowerDefaults


class CustomTensorDataset(torch.utils.data.Dataset):
    """TensorDataset with support of transforms."""

    def __init__(self, tensors, transform=None):
        assert all(tensors[0].size(0) == tensor.size(0) for tensor in tensors)
        self.tensors = tensors
        self.transform = transform

    def __getitem__(self, index):
        x = self.tensors[0][index]
        y = self.tensors[1][index]
        if self.transform:
            x = self.transform(x.numpy().astype(np.uint8))
        return x, y

    def __len__(self):
        return self.tensors[0].size(0)


class CustomNumpyDataset(torch.utils.data.Dataset):
    """NumpyDataset with support of transforms."""

    def __init__(self, path, train):
        self.path = path
        if train:
            self.tensors = (
                torch.from_numpy(np.load(os.path.join(path, "X_train.npy"))),
                torch.from_numpy(np.load(os.path.join(path, "y_train.npy")))
            )

        else:
            self.tensors = (
                torch.from_numpy(np.load(os.path.join(path, "X_test.npy"))),
                torch.from_numpy(np.load(os.path.join(path, "y_test.npy")))
            )
        self.data = self.tensors[0].squeeze().float()
        self.targets = self.tensors[-1].float()

    def __getitem__(self, index):
        x = self.tensors[0][index]
        y = self.tensors[1][index]
        return x, y

    def __len__(self):
        return self.tensors[0].size(0)


class LightningFlowerData:
    @staticmethod
    def add_data_specific_args(parent_parser):
        parser = parent_parser.add_argument_group("LightningFlowerData")
        parser.add_argument("--dataset_path", type=str, default=LightningFlowerDefaults.DATASET_FOLDER)
        parser.add_argument("--batch_size_train", type=int, default=LightningFlowerDefaults.BATCH_SIZE_TRAIN)
        parser.add_argument("--batch_size_test", type=int, default=LightningFlowerDefaults.BATCH_SIZE_TEST)
        parser.add_argument("--dataset_config", type=int, default=LightningFlowerDefaults.DATASET_CONFIG,
                            help="0 for NON-IID, 1 for IID, 2 for SAMPLED")
        parser.add_argument("--num_workers", type=int, default=LightningFlowerDefaults.NUM_WORKERS)
        return parent_parser

    @staticmethod
    def split_per_client(dataset, client_id, num_clients, shuffle=False, stratify=False):
        """Splits the dataset according to the partition size calculated with client_id/num_clients"""
        if int(client_id) >= num_clients:
            return None

        # partition size based on numer of clients
        partition_size = 1.0 / num_clients
        # indices passed to subset random sampler
        train_idx = []
        # training examples
        num_samples = len(dataset)
        # list of indices
        indices = list(range(num_samples))
        # split size based on partitions
        split = int(np.floor(partition_size * num_samples))

        # stratified split implementation
        if stratify:
            # on-demand import of sklearn lib
            from sklearn.model_selection import train_test_split
            # Split the indices in a stratified way
            train_idx, _ = train_test_split(indices,
                                            train_size=partition_size,
                                            stratify=dataset.targets,
                                            shuffle=shuffle)
            if shuffle:
                print("Client " + str(client_id) + "/" + str(num_clients) +
                      " received a random shuffled stratified split of size " + str(split))
            else:
                print("Client " + str(client_id) + "/" + str(num_clients) +
                      " received a stratified split of size " + str(split))
        else:
            if shuffle:
                np.random.shuffle(indices)
                # indices are shuffled, just use the first split
                train_idx = indices[split:]
            else:
                idx_from, idx_to = client_id * split, (client_id + 1) * split - 1
                train_idx = indices[idx_from:idx_to]
            if shuffle:
                print("Client " + str(client_id) + "/" + str(
                    num_clients) + " receives a random shuffled split of size " + str(split))
            else:
                print("Client " + str(client_id) + "/" + str(num_clients) + " receives data from index " + str(
                    idx_from) + " to " + str(idx_to))
        return SubsetRandomSampler(train_idx)



"""LightningFlower Utility"""
import inspect
import numpy as np
import os
import secrets
from pathlib import Path


def id_generator(size=6):
    """Formatted random string id generator"""
    return ''.join(str(secrets.randbelow(10)) for _ in range(size))


def printf(message):
    """Formatted printing, e.g. [filenameOfCallingFile.py]: your text"""
    print("[" + os.path.basename(inspect.stack()[1].filename) + "]: " + message)


def load_params(full_filepath):
    params = None
    # load weights
    if os.path.isfile(full_filepath):
        weights = np.load(full_filepath, allow_pickle=True)
        if weights is not None:
            params = weights.f.arr_0.item(0)
        if params is None:
            printf("Failed loading weights")
        else:
            printf("Success loading weights")
    return params


def boolean_string(s):
    """String to bool conversion"""
    if s not in {'False', 'True'}:
        raise ValueError('Not a valid boolean string')
    return s == 'True'


def create_dir(path):
    """Creating a directory, check if directory exists"""
    return Path(path).mkdir(parents=True, exist_ok=True)



"""LightningFlower Server"""
import flwr as fl
from flwr.server.client_manager import SimpleClientManager
from lightningflower.config import LightningFlowerDefaults
from lightningflower.strategy import LightningFlowerFedAvgStrategy


class LightningFlowerServer(fl.server.Server):
    @staticmethod
    def add_server_specific_args(parent_parser):
        parser = parent_parser.add_argument_group("LightningFlowerServer")
        parser.add_argument("--host_address", type=str, default=LightningFlowerDefaults.HOST_ADDRESS)
        parser.add_argument("--num_rounds", type=int, default=LightningFlowerDefaults.NR_ROUNDS)
        parser.add_argument("--max_msg_size", type=int, default=LightningFlowerDefaults.GRPC_MAX_MSG_LENGTH)
        return parent_parser

    def __init__(self, strategy=None):
        # create the client manager
        client_manager = SimpleClientManager()
        # apply default strategy
        if strategy is None:
            strategy = LightningFlowerFedAvgStrategy()
        super().__init__(client_manager=client_manager, strategy=strategy)




from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

required_packages = [
    "torch", "flwr", "pytorch-lightning", "numpy", "scikit-learn"
]

setup(
    name="lightningflower",
    version="0.1.4",
    description="Pre-packaged federated learning framework using Flower and PyTorch-Lightning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ManuelRoeder/lightningflower",
    author="Manuel Roeder",
    author_email="manuel.roeder@web.de",
    license="MIT",
    packages=["lightningflower"],
    install_requires=required_packages,
    python_requires='>=3.8.12',
    package_data={"": ["README.md", "LICENSE"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ]
)
